@@ -30,10 +30,10 @@
  * One "job" for the preadv() syscall in pagemap.c
  */
 struct page_read_iov {
-	off_t from; /* offset in pi file where to start reading from */
-	off_t end; /* the end of the read == sum to.iov_len -s */
+	off_t from;	  /* offset in pi file where to start reading from */
+	off_t end;	  /* the end of the read == sum to.iov_len -s */
 	struct iovec *to; /* destination iovs */
-	unsigned int nr; /* their number */
+	unsigned int nr;  /* their number */
 
 	struct list_head l;
 };
@@ -261,7 +261,7 @@ static int read_local_page(struct page_read *pr, unsigned long vaddr, unsigned l
 			break;
 	}
 
-	if (opts.auto_dedup) {
+	if (opts.auto_dedup && !pr->disable_dedup) {
 		ret = punch_hole(pr, pr->pi_off, len, false);
 		if (ret == -1)
 			return -1;
@@ -535,7 +535,6 @@ static int process_async_reads(struct page_read *pr)
 	fd = img_raw_fd(pr->pi);
 	list_for_each_entry_safe(piov, n, &pr->async, l) {
 		ssize_t ret;
-		off_t start = piov->from;
 		struct iovec *iovs = piov->to;
 
 		pr_debug("Read piov iovs %d, from %ju, len %ju, first %p:%zu\n", piov->nr, piov->from,
@@ -554,13 +553,16 @@ static int process_async_reads(struct page_read *pr)
 			}
 		}
 
-		if (ret != piov->end - piov->from) {
-			if (ret < 0) {
-				pr_err("Can't read async pr bytes (%zd / %ju read, %ju off, %d iovs)\n", ret,
-				       piov->end - piov->from, piov->from, piov->nr);
-				return -1;
-			}
+		if (ret < 0) {
+			pr_err("Can't read async pr bytes (%zd / %ju read, %ju off, %d iovs)\n", ret,
+			       piov->end - piov->from, piov->from, piov->nr);
+			return -1;
+		}
 
+		if (opts.auto_dedup && punch_hole(pr, piov->from, ret, false))
+			return -1;
+
+		if (ret != piov->end - piov->from) {
 			/*
 			 * The preadv() can return less than requested. It's
 			 * valid and doesn't mean error or EOF. We should advance
@@ -574,9 +576,6 @@ static int process_async_reads(struct page_read *pr)
 			goto more;
 		}
 
-		if (opts.auto_dedup && punch_hole(pr, start, ret, false))
-			return -1;
-
 		BUG_ON(pr->io_complete); /* FIXME -- implement once needed */
 
 		list_del(&piov->l);
@@ -793,6 +792,7 @@ int open_page_read_at(int dfd, unsigned long img_id, struct page_read *pr, int p
 	pr->bunch.iov_base = NULL;
 	pr->pmes = NULL;
 	pr->pieok = false;
+	pr->disable_dedup = false;
 
 	pr->pmi = open_image_at(dfd, i_typ, O_RSTR, img_id);
 	if (!pr->pmi)
@@ -853,6 +853,14 @@ int open_page_read(unsigned long img_id, struct page_read *pr, int pr_flags)
 
 #define DUP_IDS_BASE 1000
 
+void page_read_disable_dedup(struct page_read *pr)
+{
+	pr_debug("disable dedup, id: %d\n", pr->id);
+	pr->disable_dedup = true;
+	if (pr->parent)
+		page_read_disable_dedup(pr->parent);
+}
+
 void dup_page_read(struct page_read *src, struct page_read *dst)
 {
 	static int dup_ids = 1;
